﻿using HSR.CloudTaskParallelism.Client.Runtime;
using System;
using System.Collections.Generic;
using System.Threading.Tasks;

namespace HSR.CloudTaskParallelism._Samples.KnightTours {
  public class KnightTour {
    #region Initialization
    private const int _ParallelizatioDepth = 4; // depth of decision tree to create tasks for local or distributed parallelization

    private const int _Free = 0;
    private const int _First = 1;

    private static readonly int[][] _Moves = new int[][] {
      new int[] { +1, +2 },
      new int[] { +1, -2 },
      new int[] { -1, +2 },
      new int[] { -1, -2 },
      new int[] { +2, +1 },
      new int[] { +2, -1 },
      new int[] { -2, +1 },
      new int[] { -2, -1 },
    };

    private int _nofRows, _nofCols;

    public KnightTour(int nofRows, int nofCols) {
      _nofRows = nofRows;
      _nofCols = nofCols;
    }

    private int[][] _InitialBoard() {
      int[][] board = new int[_nofRows][];
      for (int x = 0; x < _nofRows; x++) {
        board[x] = new int[_nofCols];
        for (int y = 0; y < _nofCols; y++) {
          board[x][y] = _Free;
        }
      }
      return board;
    }
    #endregion

    #region Sequential Computation
    public long SequentialCountKnightTours() {
      return _CountTours(_InitialBoard(), _First, 0, 0);
    }
    #endregion

    #region Local Parallelization
    public long ParallelCountKnightTours() {
      IList<Task<long>> allTasks = new List<Task<long>>();
      _CollectParallelTasks(_InitialBoard(), allTasks, _First, 0, 0);
      Console.WriteLine("#tasks " + allTasks.Count);
      foreach (Task<long> task in allTasks) {
        task.Start();
      }
      long count = 0;
      foreach (Task<long> task in allTasks) {
        count += task.Result;
      }
      return count;
    }

    // parallel threshold must be smaller than number of fields (no solutions can be found up till level)
    private void _CollectParallelTasks(int[][] board, IList<Task<long>> allTasks, int step, int x, int y) {
      if (_InsideBoard(x, y) && board[x][y] == _Free) {
        if (step > _ParallelizatioDepth) {
          int[][] cloneBoard = _CloneBoard(board);
          allTasks.Add(new Task<long>(() => _CountTours(cloneBoard, step, x, y)));
        } else {
          board[x][y] = step;
          foreach (int[] move in _Moves) {
            _CollectParallelTasks(board, allTasks, step + 1, x + move[0], y + move[1]);
          }
          board[x][y] = _Free;
        }
      }
    }
    #endregion

    #region Distributed Parallelization
    public long DistributedCountKnightTours(Distribution distribution) {
      IList<DistributedTask<long>> allTasks = new List<DistributedTask<long>>();
      _CollectDistributedTasks(_InitialBoard(), allTasks, _First, 0, 0);
      distribution.Start(allTasks);
      long count = 0;
      foreach (DistributedTask<long> task in allTasks) {
        count += task.Result;
      }
      return count;
    }

    private void _CollectDistributedTasks(int[][] board, IList<DistributedTask<long>> allTasks, int step, int x, int y) {
      if (_InsideBoard(x, y) && board[x][y] == _Free) {
        if (step > _ParallelizatioDepth) {
          int[][] cloneBoard = _CloneBoard(board);
          allTasks.Add(DistributedTask.New(() => _CountTours(cloneBoard, step, x, y)));
        } else {
          board[x][y] = step;
          foreach (int[] move in _Moves) {
            _CollectDistributedTasks(board, allTasks, step + 1, x + move[0], y + move[1]);
          }
          board[x][y] = _Free;
        }
      }
    }
    #endregion

    #region Task Code 
    private long _CountTours(int[][] board, int step, int x, int y) {
      long count = 0;
      if (_InsideBoard(x, y)) {
        if (step > _nofRows * _nofCols) {
          if (board[x][y] == _First) {
            count = 1;
          }
        } else if (board[x][y] == _Free && !_IsolatedField(board, x, y)) {
          board[x][y] = step;
          foreach (int[] move in _Moves) {
            count += _CountTours(board, step + 1, x + move[0], y + move[1]);
          }
          board[x][y] = _Free;
        }
      }
      return count;
    }

    private int[][] _CloneBoard(int[][] board) {
      int[][] copy = new int[_nofRows][];
      for (int x = 0; x < _nofRows; x++) {
        copy[x] = new int[_nofCols];
        for (int y = 0; y < _nofCols; y++) {
          copy[x][y] = board[x][y];
        }
      }
      return copy;
    }

    private bool _InsideBoard(int x, int y) {
      return x >= 0 && x < _nofRows && y >= 0 && y < _nofCols;
    }

    // Pruning heuristics
    private bool _IsolatedField(int[][] board, int currentX, int currentY) {
      for (int x = 0; x < _nofRows; x++) {
        for (int y = 0; y < _nofCols; y++) {
          if (x != currentX && y != currentY && board[x][y] == _Free && _CountFreeNeighbors(board, x, y) == 0) {
            return true;
          }
        }
      }
      return false;
    }

    private int _CountFreeNeighbors(int[][] board, int currentX, int currentY) {
      int count = 0;
      for (int k = 0; k < _Moves.Length; k++) {
        int neighborX = currentX + _Moves[k][0];
        int neighborY = currentY + _Moves[k][1];
        if (_InsideBoard(neighborX, neighborY) && board[neighborX][neighborY] == _Free) {
          count++;
        }
      }
      return count;
    }
    #endregion
  }
}
