﻿using HSR.CloudTaskParallelism.Client.Runtime;
using System;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Windows.Media.Imaging;

namespace HSR.CloudTaskParallelism._Samples.Mandelbrot {
  public partial class MainWindow : Window {
    private Distribution _distribution = new Distribution(AppSettings.Default.TaskExecutionServiceUri, AppSettings.Default.TaskExecutionAuthorizationToken);
        
    private const double _Left = -2.2;
    private const double _Top = -1.13;
    private const double _Width = 2.9;
    private const double _Height = 2.27;
    
    public MainWindow() {
      InitializeComponent();
    }

    private async void computeButton_Click(object sender, RoutedEventArgs e) {
      IsEnabled = false;
      int horizontalPixels = int.Parse(widthTextBox.Text);
      int verticalPixels = int.Parse(heightTextBox.Text);
      await _RunComputation("Distributed", () => Mandelbrot.DistributedCompute(_distribution, _Left, _Top, _Width, _Height, horizontalPixels, verticalPixels));
      await _RunComputation("Parallel", () => Mandelbrot.ParallelCompute(_Left, _Top, _Width, _Height, horizontalPixels, verticalPixels));
      await _RunComputation("Sequential", () => Mandelbrot.SequentialCompute(_Left, _Top, _Width, _Height, horizontalPixels, verticalPixels));
      IsEnabled = true;
    }

    private async Task _RunComputation(string name, Func<byte[][]> calculation) {
      byte[][] pixels = null; // not nice: because of async/await model (need to keep UI-updates in UI scheduler but run time-consuming activity in TPL threadpool)
      double totalMillis = 0;
      await Task.Run(() => {
        DateTime startTime = DateTime.Now;
        pixels = calculation();
        DateTime stopTime = DateTime.Now; 
        totalMillis = (stopTime - startTime).TotalMilliseconds;
      });
      string title = string.Format("{0} {1} ms", name, totalMillis);
      _ShowPixels(title, pixels);
    }

    private void _ShowPixels(string title, byte[][] pixels) {
      BitmapView view = new BitmapView();
      view.Title = title;
      view.mandelbrotImage.Source = _GetBitmap(pixels);
      view.Show();
    }

    private BitmapSource _GetBitmap(byte[][] pixels) {
      int bitmapHeight = pixels.Length;
      int bitmapWidth = pixels[0].Length; // TODO: Use multi-dimensional array
      WriteableBitmap bitmap = new WriteableBitmap(bitmapWidth, bitmapHeight, 96, 96, PixelFormats.Bgra32, null);
      bitmap.Lock();
      unsafe {
        byte* buffer = (byte*)bitmap.BackBuffer.ToPointer();
        for (int row = 0; row < bitmapHeight; row++) {
          for (int col = 0; col < bitmapWidth; col++) {
            Color color = _GetColor(pixels[row][col]);
            *buffer++ = color.B;
            *buffer++ = color.G;
            *buffer++ = color.R;
            *buffer++ = color.A;
          }
        }
      }
      bitmap.AddDirtyRect(new Int32Rect(0, 0, bitmapWidth, bitmapHeight));
      bitmap.Unlock();
      return bitmap;
    }

    private Color _GetColor(byte value) {
      return Color.FromRgb((byte)(value * 10 % 256), (byte)(value * 20 % 256), (byte)(value * 30 % 256));
    }
  }
}
